@@ -36,6 +36,7 @@ Shortcuts:
   F                         find
   Q                         quit
   P                         print
+  T                         toggle show/hide toolbar
   Escape                    halt animation
   Ctrl-drag                 zoom in/out
   Shift-drag                zooms an area
@@ -56,11 +57,15 @@ Shortcuts:
         '-g', '--geometry',
         action='store', dest='geometry',
         help='default window size in form WxH')
+    parser.add_argument(
+        '--hide-toolbar',
+        action='store_true', dest='hide_toolbar',
+        help='Hides the toolbar on start.')
 
     options = parser.parse_args()
     inputfile = options.inputfile
 
-    width = height = 512
+    width, height = 610, 610
     if options.geometry:
         try:
             width, height = (int(i) for i in options.geometry.split('x'))
@@ -76,6 +81,9 @@ Shortcuts:
         else:
             win.open_file(inputfile)
 
+    if options.hide_toolbar:
+        win.uimanager.get_widget('/ToolBar').set_visible(False)
+
     if sys.platform != 'win32':
         # Reset KeyboardInterrupt SIGINT handler, so that glib loop can be stopped by it
         import signal
