@@ -15,6 +15,7 @@
 #
 import math
 import operator
+import warnings
 
 import gi
 gi.require_version('Gtk', '3.0')
@@ -22,10 +23,12 @@ gi.require_version('PangoCairo', '1.0')
 
 from gi.repository import GObject
 from gi.repository import Gdk
+from gi.repository import Gtk
 from gi.repository import GdkPixbuf
 from gi.repository import Pango
 from gi.repository import PangoCairo
 import cairo
+import numpy
 
 _inf = float('inf')
 _get_bounding = operator.attrgetter('bounding')
@@ -68,6 +71,9 @@ class Shape:
     def search_text(self, regexp):
         return False
 
+    def get_smallest_distance(self, x, y):
+        return None
+
     @staticmethod
     def _bounds_from_points(points):
         x0, y0 = points[0]
@@ -88,10 +94,14 @@ class Shape:
                 ya, yb = min(ya, y0), max(yb, y1)
         return xa, ya, xb, yb
 
+    def get_text(self):
+        return None
+
 
 class TextShape(Shape):
 
     LEFT, CENTER, RIGHT = -1, 0, 1
+    DEFAULT_FONTNAME = Gtk.Settings.get_default().get_property("gtk-font-name")
 
     def __init__(self, pen, x, y, j, w, t):
         Shape.__init__(self)
@@ -102,6 +112,10 @@ class TextShape(Shape):
         self.w = w  # width
         self.t = t  # text
 
+    def _font_available(self, fontname, pango_context):
+        available_fonts = [family.get_name() for family in pango_context.list_families()]
+        return fontname in available_fonts
+
     def _draw(self, cr, highlight, bounding):
 
         try:
@@ -150,7 +164,16 @@ class TextShape(Shape):
             assert success
             layout.set_attributes(attrs)
 
-            font.set_family(self.pen.fontname)
+            if self._font_available(self.pen.fontname, pango_context=context):
+                font.set_family(self.pen.fontname)
+            else:
+                msg = "Font family {fontname!r} is not available, using {default!r}".format(
+                    fontname=self.pen.fontname,
+                    default=self.DEFAULT_FONTNAME
+                )
+                warnings.warn(msg)
+                font.set_family(self.DEFAULT_FONTNAME)
+
             font.set_absolute_size(self.pen.fontsize*Pango.SCALE)
             layout.set_font_description(font)
 
@@ -207,6 +230,9 @@ class TextShape(Shape):
         x, w, j = self.x, self.w, self.j
         return x - 0.5 * (1 + j) * w, -_inf, x + 0.5 * (1 - j) * w, _inf
 
+    def get_text(self):
+        return self.t
+
 
 class ImageShape(Shape):
 
@@ -408,6 +434,67 @@ class BezierShape(Shape):
             cr.set_source_rgba(*pen.color)
             cr.stroke()
 
+    def get_smallest_distance(self, x, y):
+        min_squared_distance = float('inf')
+
+        points_iter = iter(self.points)
+        x0, y0 = next(points_iter)
+
+        while True:
+            try:
+                x1, y1 = next(points_iter)
+
+            except StopIteration:
+                break
+
+            x2, y2 = next(points_iter)
+            x3, y3 = next(points_iter)
+
+            _e1 = -5
+            _e2 = (x0 - 3 * x1 + 3 * x2 - x3)
+            _e3 = (y0 - 3 * y1 + 3 * y2 - y3)
+            _e4 = 2 * x1
+            _e5 = 2 * y1
+            _e6 = x0**2
+            _e7 = y0**2
+            _e8 = -2
+            _e9 = (x0 - _e4 + x2)
+            _e10 = (y0 - _e5 + y2)
+            _e11 = 2 * x0
+            _e12 = 2 * y0
+            _e13 = 5 * _e6
+            _e14 = 5 * _e7
+            _e15 = x1**2
+            _e16 = y1**2
+            coefficients = [
+                (x - x0) * (x0 - x1) + (y - y0) * (y0 - y1),
+                _e13 + 3 * _e15 + _e11 * (_e1 * x1 + x2) - 2 * x * _e9 + _e14 + 3 * _e16 + _e12 * (_e1 * y1 + y2) - 2 * y * _e10,
+                -10 * _e6 + 9 * x1 * (_e8 * x1 + x2) + x * _e2 + x0 * (30 * x1 - 12 * x2 + x3) - 10 * _e7 + 9 * y1 * (_e8 * y1 + y2) + y * _e3 + y0 * (30 * y1 - 12 * y2 + y3),
+                2 * (_e13 + 18 * _e15 + 3 * x2**2 + _e4 * (-9 * x2 + x3) - _e11 * (10 * x1 - 6 * x2 + x3) + _e14 + 3 * (6 * _e16 - 6 * y1 * y2 + y2**2) + _e5 * y3 - _e12 * (10 * y1 - 6 * y2 + y3)),
+                _e1 * _e9 * _e2 - 5 * _e10 * _e3,
+                _e2**2 + _e3**2
+            ]
+            coefficients.reverse()
+
+            for t in numpy.roots(coefficients):
+                if 1e-6 < abs(t.imag):
+                    continue
+
+                t = t.real
+                if t < 0:
+                    t = 0
+                elif 1 < t:
+                    t = 1
+
+                squared_distance = ((1 - t)**3 * x0 + 3 * (1 - t)**2 * t * x1 + 3 * (1 - t) * t**2 * x2 + t**3 * x3 - x)**2 + ((1 - t)**3 * y0 + 3 * (1 - t)**2 * t * y1 + 3 * (1 - t) * t**2 * y2 + t**3 * y3 - y)**2
+                if squared_distance < min_squared_distance:
+                    min_squared_distance = squared_distance
+
+            x0 = x3
+            y0 = y3
+
+        return math.sqrt(min_squared_distance)
+
 
 class CompoundShape(Shape):
 
@@ -429,6 +516,13 @@ class CompoundShape(Shape):
                 return True
         return False
 
+    def get_text(self):
+        for shape in self.shapes:
+            text = shape.get_text()
+            if text is not None:
+                return text
+        return None
+
 
 class Url(object):
 
@@ -469,7 +563,7 @@ class Element(CompoundShape):
 
 class Node(Element):
 
-    def __init__(self, id, x, y, w, h, shapes, url):
+    def __init__(self, id, x, y, w, h, shapes, url, tooltip):
         Element.__init__(self, shapes)
 
         self.id = id
@@ -482,6 +576,7 @@ class Node(Element):
         self.y2 = y + 0.5*h
 
         self.url = url
+        self.tooltip = tooltip
 
     def is_inside(self, x, y):
         return self.x1 <= x and x <= self.x2 and self.y1 <= y and y <= self.y2
@@ -510,11 +605,12 @@ def square_distance(x1, y1, x2, y2):
 
 class Edge(Element):
 
-    def __init__(self, src, dst, points, shapes):
+    def __init__(self, src, dst, points, shapes, tooltip):
         Element.__init__(self, shapes)
         self.src = src
         self.dst = dst
         self.points = points
+        self.tooltip = tooltip
 
     RADIUS = 10
 
@@ -532,10 +628,19 @@ class Edge(Element):
         return False
 
     def get_jump(self, x, y):
-        if self.is_inside_begin(x, y):
-            return Jump(self, self.dst.x, self.dst.y, highlight=set([self, self.dst]))
-        if self.is_inside_end(x, y):
-            return Jump(self, self.src.x, self.src.y, highlight=set([self, self.src]))
+        for shape in self.shapes:
+            x1, y1, x2, y2 = shape.bounding
+            if (x1 - self.RADIUS) <= x and x <= (x2 + self.RADIUS) and (y1 - self.RADIUS) <= y and y <= (y2 + self.RADIUS):
+                break
+
+        else:
+            return None
+
+        for shape in self.shapes:
+            distance = shape.get_smallest_distance(x, y)
+            if distance is not None and distance <= self.RADIUS:
+                return Jump(self, self.src.x, self.src.y)
+
         return None
 
     def __repr__(self):
@@ -562,15 +667,23 @@ class Graph(Shape):
     def get_size(self):
         return self.width, self.height
 
-    def _draw_shapes(self, cr, bounding):
+    def _draw_shapes(self, cr, bounding, highlight_items):
         for shape in self.shapes:
             if bounding is None or shape._intersects(bounding):
-                shape._draw(cr, highlight=False, bounding=bounding)
+                shape._draw(cr, highlight=(shape in highlight_items), bounding=bounding)
 
     def _draw_nodes(self, cr, bounding, highlight_items):
+        highlight_nodes = []
+        for element in highlight_items:
+            if isinstance(element, Edge):
+                highlight_nodes.append(element.src)
+                highlight_nodes.append(element.dst)
+            else:
+                highlight_nodes.append(element)
+
         for node in self.nodes:
             if bounding is None or node._intersects(bounding):
-                node._draw(cr, highlight=(node in highlight_items), bounding=bounding)
+                node._draw(cr, highlight=(node in highlight_nodes), bounding=bounding)
 
     def _draw_edges(self, cr, bounding, highlight_items):
         for edge in self.edges:
@@ -593,7 +706,8 @@ class Graph(Shape):
         cr.set_line_cap(cairo.LINE_CAP_BUTT)
         cr.set_line_join(cairo.LINE_JOIN_MITER)
 
-        self._draw_shapes(cr, bounding)
+        self._draw_shapes(cr, bounding, highlight_items)
+
         if self.outputorder == 'edgesfirst':
             self._draw_edges(cr, bounding, highlight_items)
             self._draw_nodes(cr, bounding, highlight_items)
